import logging
import os
import sys

from loguru import logger

from planqk.commons.constants import LOG_LEVEL_ENV, DEFAULT_LOG_LEVEL


class LogHandler(logging.Handler):
    """
    A handler class which allows for the use of Loguru with the standard logging module.
    """

    def __init__(self, name: str = None) -> None:
        if name is not None:
            self._logger = logger.patch(lambda record: record.update(name=name))
        else:
            self._logger = logger
        super().__init__()

    def emit(self, record):
        # Get corresponding Loguru level if it exists
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        self._logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )


def init_logging():
    """
    Initialize the logging configuration. The logging level is set to the value of the LOG_LEVEL environment variable.
    """
    logging_level = os.environ.get(LOG_LEVEL_ENV, DEFAULT_LOG_LEVEL).upper()

    logging.getLogger().handlers = [LogHandler()]
    logging.getLogger().setLevel(logging_level)

    logger.configure(handlers=[{"sink": sys.stdout, "level": logging_level}])

    for stdlib_logger, loguru_handler in (
        (logging.getLogger(name), LogHandler(name))
        for name in logging.root.manager.loggerDict
    ):
        stdlib_logger.handlers = [loguru_handler]
        stdlib_logger.propagate = False



import os
from typing import Dict


class DataPool:
    """
    Runtime abstraction for a mounted data pool at /var/runtime/datapool.
    """

    def __init__(self, path: str):
        if not os.path.exists(path):
            raise FileNotFoundError(f"Path \"{path}\" does not exist")

        if not os.path.isdir(path):
            raise NotADirectoryError(f"Path \"{path}\" is not a directory")

        self._path = path
        self._name = os.path.basename(path)

    @property
    def name(self) -> str:
        """
        Returns the name of the data pool.
        """
        return self._name

    @property
    def path(self) -> str:
        """
        Returns the path to the data pool.
        """
        return self._path

    def list_files(self) -> Dict[str, str]:
        """
        List all files in the data pool directory.
        """
        files = {}

        for file_name in os.listdir(self._path):
            file_path = os.path.join(self._path, file_name)
            if os.path.isfile(file_path):
                files[file_name] = file_path

        return files

    def open(self, name: str, mode: str = 'r'):
        """
        Open a file from the data pool by its name, supporting context manager usage.
        Usage: with data_pool.open("filename") as f:
        """
        file_path = os.path.join(self._path, name)

        if not os.path.isfile(file_path):
            raise FileNotFoundError(f"File \"{name}\" not found in data pool")

        return open(file_path, mode)



import os
from typing import ByteString, Union, Any

from planqk.commons.constants import OUTPUT_DIRECTORY_ENV, DEFAULT_OUTPUT_DIRECTORY
from planqk.commons.file import write_str_to_file, write_blob_to_file
from planqk.commons.json import any_to_json


def write_string_output(file_name: str, content: str):
    """
    Write a string to a file in the output directory.

    :param file_name: The name of the file.
    :param content: The content to write to the file.
    """
    directory_path = os.environ.get(OUTPUT_DIRECTORY_ENV, DEFAULT_OUTPUT_DIRECTORY)
    write_str_to_file(directory_path, file_name, content)


def write_binary_output(file_name: str, content: Union[bytes, bytearray, ByteString]):
    """
    Write a binary blob to a file in the output directory.

    :param file_name: The name of the file.
    :param content: The content to write to the file.
    """
    directory_path = os.environ.get(OUTPUT_DIRECTORY_ENV, DEFAULT_OUTPUT_DIRECTORY)
    write_blob_to_file(directory_path, file_name, content)


def print_output(content: Any):
    """
    Print the content as JSON. Wraps the output in a PlanQK:Job:MultilineResult block.

    :param content: The content to print. Must be JSON serializable.
    """
    output = any_to_json(content)

    if output:
        print(f"PlanQK:Job:MultilineResult\n{output}\nPlanQK:Job:MultilineResult")



import os
from typing import Dict, Any

import yaml


def read_to_dict(file_path: str) -> Dict[str, Any]:
    """
    Read a YAML file and return it as a dictionary.

    :param file_path: The path to the file
    :return: The content of the file as a dictionary
    """
    with open(file_path, "r") as file:
        openapi = yaml.safe_load(file)

    return openapi


def write_dict(directory_path: str, openapi: Dict[str, Any], filename: str = "openapi.yaml"):
    """
    Write a dictionary to a YAML file.

    :param directory_path: The directory where the file should be saved
    :param openapi: The dictionary to be written to the file
    :param filename: The name of the file, defaults to "openapi.yaml"
    """
    file_path = os.path.join(directory_path, filename)
    with open(file_path, "w") as file:
        yaml.dump(openapi, file, sort_keys=False)



import os
from typing import List

from planqk.commons.constants import INPUT_DIRECTORY_ENV, DEFAULT_INPUT_DIRECTORY
from planqk.commons.file import list_directory_files


def read_input_as_string(file_or_attribute: str, input_directories: List[str] = None) -> str:
    """
    Read the content of a file from the input directory as a string.

    :param file_or_attribute: The name of the file or attribute (API) to read.
    :param input_directories: The directories to search for the file. If None, the environment variable INPUT_DIRECTORY is used.
    :return: The content of the file as a string.
    """
    if input_directories is None:
        input_directory = os.environ.get(INPUT_DIRECTORY_ENV, DEFAULT_INPUT_DIRECTORY)
        input_directories = [input_directory]

    # find first input directory that exists
    for input_directory in input_directories:
        if os.path.exists(input_directory):
            input_directory = os.path.abspath(input_directory)
            break
    else:
        raise FileNotFoundError(f"None of the input directories exist: {input_directories}")

    # remove possible file extension
    name = file_or_attribute.split(".")[0]

    input_files = list_directory_files(input_directory)

    if name not in input_files:
        raise FileNotFoundError(f"File '{name}' not found in input directory '{input_directory}'")

    with open(input_files[name], "r") as file:
        file_content = file.read()

    return file_content



import os
from typing import Dict, Any

import yaml

from planqk.commons.constants import ENTRYPOINT_ENV, DEFAULT_ENTRYPOINT
from planqk.commons.openapi.schema import generate_parameter_schema, generate_return_schema
from planqk.commons.openapi.template import get_template_managed_service
from planqk.commons.reflection import resolve_signature


def generate_openapi(entrypoint: str = os.environ.get(ENTRYPOINT_ENV, DEFAULT_ENTRYPOINT),
                     title: str = "PLANQK Service API",
                     version: str = "1") -> None:
    """
    Generate an OpenAPI specification for a PLANQK Service.

    :param entrypoint: The entrypoint of the service
    :param title: The title of the service, defaults to "PLANQK Service API"
    :param version: The version of the service, defaults to "1"
    :return:
    """
    entrypoint_signature = resolve_signature(entrypoint)

    parameter_schemas, parameter_schema_definitions = generate_parameter_schema(entrypoint_signature)
    return_schema, return_schema_definitions = generate_return_schema(entrypoint_signature)

    schema_definitions = {}
    schema_definitions.update(parameter_schema_definitions)
    schema_definitions.update(return_schema_definitions)

    openapi_template = get_template_managed_service()

    openapi = populate_openapi_template(openapi_template, parameter_schemas, return_schema, schema_definitions, title, version)

    print(yaml.dump(openapi, sort_keys=False))


def populate_openapi_template(openapi_template: Dict[str, Any],
                              input_schema: Dict[str, Any],
                              output_schema: Dict[str, Any],
                              schema_definitions: Dict[str, Any],
                              title: str,
                              version: str = "1") -> Dict[str, Any]:
    # deep copy of openapi_template
    openapi = dict(openapi_template)

    openapi["components"]["schemas"].update(schema_definitions)

    # start execution route
    if input_schema:
        openapi["paths"]["/"]["post"]["requestBody"]["content"]["application/json"]["schema"]["properties"] = input_schema
    else:
        del openapi["paths"]["/"]["post"]["requestBody"]

    # result route
    if "properties" in output_schema:
        (openapi["paths"]["/{id}/result"]["get"]["responses"]["200"]["content"]["application/json"]["schema"]["properties"]
         .update(output_schema["properties"]))

    openapi["info"]["title"] = title
    openapi["info"]["version"] = version

    return openapi



"""
Parameter mapping utilities for PLANQK runtime system.

This module provides functions to map input files to function parameters,
including support for DataPool parameter injection.
"""
import json
import os
from inspect import Signature, Parameter
from typing import Any, Dict, get_origin, get_args, Union, Optional

from loguru import logger
from pydantic import BaseModel

from planqk.commons.datapool import DataPool


def files_to_parameters(input_files: Dict[str, str], signature: Signature) -> Dict[str, Any]:
    """
    Maps input files to parameters of a function signature. If a parameter is not found in the input files,
    the default value of the parameter is used. If the parameter is optional and has no default value,
    it is set to None.

    DataPool parameters are automatically injected based on parameter name and type annotation.

    :param input_files: The input files to be mapped to the parameters.
                        The keys are the parameter names. The values are the file paths.
    :param signature: The signature of the function to which the input files should be mapped.
    :return: A dictionary containing the parameters of the function with input files mapped to them.
    """
    parameters = {}

    for parameter in signature.parameters.values():
        parameter_name = parameter.name
        parameter_type = parameter.annotation

        # Check if parameter is DataPool type
        if is_datapool_type(parameter_type):
            try:
                datapool_instance = create_datapool_instance(parameter_name)
                parameters[parameter_name] = datapool_instance
            except (FileNotFoundError, NotADirectoryError) as e:
                logger.warning(
                    f"DataPool directory not found for parameter '{parameter_name}': {e}"
                )
                if is_optional_type(parameter_type):
                    parameters[parameter_name] = None
                else:
                    # For required DataPool parameters, we might want to fail fast
                    raise
            continue

        if parameter_name not in input_files:
            try:
                parameters[parameter_name] = map_default(parameter)
            except TypeError:
                pass
            continue

        parameters[parameter_name] = map_input_file(input_files[parameter_name], parameter)

    return parameters


def map_input_file(input_file: str, parameter: Parameter) -> Any:
    """
    Maps an input file to a parameter of a function signature.

    :param input_file: The path to the input file.
    :param parameter: The parameter to which the input file should be mapped.
    :return: The value of the input file as the type of the parameter.
    """
    parameter_type = parameter.annotation

    origin = get_origin(parameter_type)
    if origin:
        parameter_type = origin

    with open(input_file, "r", encoding="utf-8") as file:
        file_content = file.read()

    return str_to_parameter_type(file_content, parameter_type)


def map_default(parameter: Parameter) -> Any:
    """
    Maps the default value of a parameter to the parameter type.

    :param parameter: The parameter for which the default value should be mapped.
    :return: The default value of the parameter.
    """
    parameter_type = parameter.annotation

    # use default value if available
    if parameter.default != parameter.empty:
        return parameter.default

    # if optional w/o default, set to None
    if is_optional_type(parameter_type):
        return None

    # if it is a Pydantic model, try to create an empty instance (using defaults if available)
    if issubclass(parameter_type, BaseModel):
        return parameter_type.model_validate_json("{}")

    raise TypeError(f"Could not find a default value for parameter '{parameter.name}' of type '{parameter_type}'")


def str_to_parameter_type(data: str, parameter_type: Any) -> Any:
    """
    Converts a string to a parameter type.

    :param data: The string to be converted.
    :param parameter_type: The type to which the string should be converted.
    :return: The value of the string as the type of the parameter.
    """
    if issubclass(parameter_type, str):
        return data

    if issubclass(parameter_type, bool):
        return bool(data)

    if issubclass(parameter_type, int):
        return int(data)

    if issubclass(parameter_type, float):
        return float(data)

    if issubclass(parameter_type, (list, dict)):
        return json.loads(data)

    if issubclass(parameter_type, BaseModel):
        return parameter_type.model_validate_json(data)

    raise ValueError(f"Type {parameter_type} is not supported")


def is_simple_type(value: Any) -> bool:
    """
    Checks if the value is a simple type (str, int, float, bool).
    """
    return isinstance(value, (str, int, float, bool))


def is_container_type(value: Any) -> bool:
    """
    Checks if the value is a container type (list, tuple, Union, Optional).
    """
    return value is list or value is tuple or value is Union or value is Optional


def is_optional_type(value: Any) -> bool:
    """
    Checks if the value is an optional type (Union with None or Optional).
    """
    origin = get_origin(value)
    if origin is Union:
        args = get_args(value)
        return len(args) == 2 and type(None) in args
    return origin is Optional


def is_datapool_type(parameter_type: Any) -> bool:
    """
    Check if parameter type is DataPool or Optional[DataPool].

    :param parameter_type: The parameter type annotation to check
    :return: True if the parameter type is DataPool or Optional[DataPool], False otherwise
    """
    # Handle Optional[DataPool] case
    origin = get_origin(parameter_type)
    if origin is Union:
        args = get_args(parameter_type)
        return any(arg is DataPool for arg in args if arg is not type(None))

    return parameter_type is DataPool


def create_datapool_instance(parameter_name: str) -> DataPool:
    """
    Create DataPool instance for parameter based on parameter name.

    Creates a DataPool instance pointing to '/var/runtime/datapool/{parameter_name}'.

    :param parameter_name: The name of the parameter, used as the datapool directory name
    :return: DataPool instance for the specified parameter
    :raises FileNotFoundError: If the datapool directory doesn't exist
    :raises NotADirectoryError: If the datapool path exists but is not a directory
    """
    # Import here to get the current value, allowing tests to mock it
    from planqk.commons.constants import DEFAULT_DATAPOOL_DIRECTORY

    datapool_path = os.path.join(DEFAULT_DATAPOOL_DIRECTORY, parameter_name)
    return DataPool(datapool_path)



import json
from typing import Optional

from pydantic import BaseModel


def any_to_json(data: any) -> Optional[str]:
    """
    Convert any data to a JSON string, if JSON serializable. If the data is a Pydantic
    model, the model_dump_json method is used.

    :param data: The data to convert.
    :return: The JSON string representation of the data.
    """
    if data is None:
        return None

    if issubclass(data.__class__, BaseModel):
        return data.model_dump_json(indent=2)

    return json.dumps(data, default=lambda o: getattr(o, "__dict__", str(o)), indent=2)



LOG_LEVEL_ENV = "LOG_LEVEL"
DEFAULT_LOG_LEVEL = "INFO"

ENTRYPOINT_ENV = "ENTRYPOINT"
DEFAULT_ENTRYPOINT = "user_code.src.program:run"

INPUT_DIRECTORY_ENV = "INPUT_DIRECTORY"
DEFAULT_INPUT_DIRECTORY = "/var/runtime/input"

OUTPUT_DIRECTORY_ENV = "OUTPUT_DIRECTORY"
DEFAULT_OUTPUT_DIRECTORY = "/var/runtime/output"

WORKSPACE_ENV = "WORKSPACE"
DEFAULT_WORKSPACE = "./user_code"

DATAPOOL_DIRECTORY_ENV = "DATAPOOL_DIRECTORY"
DEFAULT_DATAPOOL_DIRECTORY = "/var/runtime/datapool"
